@@ -37,6 +37,7 @@ namespace xt
         using inner_strides_type = strides_type;
         using inner_backstrides_type = backstrides_type;
         using temporary_type = xtensor_container<EC, N, L>;
+        static constexpr layout_type layout = L;
     };
 
     template <class EC, std::size_t N, layout_type L>
@@ -59,13 +60,13 @@ namespace xt
      * @sa xtensor
      */
     template <class EC, size_t N, layout_type L>
-    class xtensor_container : public xstrided_container<xtensor_container<EC, N, L>, L>,
+    class xtensor_container : public xstrided_container<xtensor_container<EC, N, L>>,
                               public xcontainer_semantic<xtensor_container<EC, N, L>>
     {
     public:
 
         using self_type = xtensor_container<EC, N, L>;
-        using base_type = xstrided_container<self_type, L>;
+        using base_type = xstrided_container<self_type>;
         using semantic_base = xcontainer_semantic<self_type>;
         using container_type = typename base_type::container_type;
         using value_type = typename base_type::value_type;
@@ -76,6 +77,7 @@ namespace xt
         using shape_type = typename base_type::shape_type;
         using inner_shape_type = typename base_type::inner_shape_type;
         using strides_type = typename base_type::strides_type;
+        using backstrides_type = typename base_type::backstrides_type;
         using inner_strides_type = typename base_type::inner_strides_type;
 
         xtensor_container();
@@ -131,6 +133,7 @@ namespace xt
         using inner_strides_type = strides_type;
         using inner_backstrides_type = backstrides_type;
         using temporary_type = xtensor_container<EC, N, L>;
+        static constexpr layout_type layout = L;
     };
 
     template <class EC, std::size_t N, layout_type L>
@@ -154,17 +157,18 @@ namespace xt
      * @tparam L The layout_type of the adaptor.
      */
     template <class EC, std::size_t N, layout_type L>
-    class xtensor_adaptor : public xstrided_container<xtensor_adaptor<EC, N, L>, L>,
+    class xtensor_adaptor : public xstrided_container<xtensor_adaptor<EC, N, L>>,
                             public xadaptor_semantic<xtensor_adaptor<EC, N, L>>
     {
     public:
 
         using self_type = xtensor_adaptor<EC, N, L>;
-        using base_type = xstrided_container<self_type, L>;
+        using base_type = xstrided_container<self_type>;
         using semantic_base = xadaptor_semantic<self_type>;
         using container_type = typename base_type::container_type;
         using shape_type = typename base_type::shape_type;
         using strides_type = typename base_type::strides_type;
+        using backstrides_type = typename base_type::backstrides_type;
 
         using container_closure_type = adaptor_closure_t<container_type>;
 
@@ -191,7 +195,7 @@ namespace xt
         const container_type& data_impl() const noexcept;
 
         using temporary_type = typename xcontainer_inner_types<self_type>::temporary_type;
-        void assign_temporary_impl(temporary_type& tmp);
+        void assign_temporary_impl(temporary_type&& tmp);
 
         friend class xcontainer<xtensor_adaptor<EC, N, L>>;
         friend class xadaptor_semantic<xtensor_adaptor<EC, N, L>>;
@@ -222,7 +226,7 @@ namespace xt
         : base_type()
     {
         base_type::reshape(xt::shape<shape_type>(t), true);
-        L == layout_type::row_major ? nested_copy(m_data.begin(), t) : nested_copy(this->xbegin(), t);
+        L == layout_type::row_major ? nested_copy(m_data.begin(), t) : nested_copy(this->template begin<layout_type::row_major>(), t);
     }
 
     /**
@@ -319,8 +323,10 @@ namespace xt
         : base_type()
     {
         // Avoids unintialized data because of (m_shape == shape) condition
-        // in reshape (called by assign), which is always true when dimension == 0.
-        if (e.derived_cast().dimension() == 0)
+        // in reshape (called by assign), which is always true when size() == 1.
+        // The condition dimension() == 0 as in xarray is not sufficient because
+        // the shape is always initialized since it has a static number of dimensions.
+        if (e.derived_cast().size() == 1)
         {
             m_data.resize(1);
         }
@@ -441,14 +447,12 @@ namespace xt
     }
 
     template <class EC, std::size_t N, layout_type L>
-    inline void xtensor_adaptor<EC, N, L>::assign_temporary_impl(temporary_type& tmp)
+    inline void xtensor_adaptor<EC, N, L>::assign_temporary_impl(temporary_type&& tmp)
     {
-        // TODO (performance improvement) : consider moving tmps shape and strides
-        base_type::shape_impl() = tmp.shape();
-        base_type::strides_impl() = tmp.strides();
-        base_type::backstrides_impl() = tmp.backstrides();
-        m_data.resize(tmp.size());
-        std::copy(tmp.data().cbegin(), tmp.data().cend(), m_data.begin());
+        base_type::shape_impl() = std::move(const_cast<shape_type&>(tmp.shape()));
+        base_type::strides_impl() = std::move(const_cast<strides_type&>(tmp.strides()));
+        base_type::backstrides_impl() = std::move(const_cast<backstrides_type&>(tmp.backstrides()));
+        m_data = std::move(tmp.data());
     }
 }
 
